@@ -6,26 +6,24 @@ Gives a multi-value dictionary object (MultiDict) plus several wrappers
 import cgi
 import copy
 import sys
-from webob.util.dictmixin import DictMixin
-try:
-    reversed
-except NameError:
-    from webob.util.reversed import reversed
 
-__all__ = ['MultiDict', 'UnicodeMultiDict', 'NestedMultiDict', 'NoVars']
+from UserDict import DictMixin
+
+
+
+__all__ = ['MultiDict', 'UnicodeMultiDict', 'NestedMultiDict', 'NoVars',
+           'TrackableMultiDict']
 
 class MultiDict(DictMixin):
-
     """
-    An ordered dictionary that can have multiple values for each key.
-    Adds the methods getall, getone, mixed, and add to the normal
-    dictionary interface.
+        An ordered dictionary that can have multiple values for each key.
+        Adds the methods getall, getone, mixed and extend and add to the normal
+        dictionary interface.
     """
 
     def __init__(self, *args, **kw):
         if len(args) > 1:
-            raise TypeError(
-                "MultiDict can only be called with one positional argument")
+            raise TypeError("MultiDict can only be called with one positional argument")
         if args:
             if hasattr(args[0], 'iteritems'):
                 items = list(args[0].iteritems())
@@ -36,9 +34,10 @@ class MultiDict(DictMixin):
             self._items = items
         else:
             self._items = []
-        self._items.extend(kw.iteritems())
+        if kw:
+            self._items.extend(kw.iteritems())
 
-    #@classmethod
+    @classmethod
     def view_list(cls, lst):
         """
         Create a dict that is a view on the given list
@@ -51,25 +50,20 @@ class MultiDict(DictMixin):
         obj._items = lst
         return obj
 
-    view_list = classmethod(view_list)
-
-    #@classmethod
+    @classmethod
     def from_fieldstorage(cls, fs):
         """
         Create a dict from a cgi.FieldStorage instance
         """
         obj = cls()
-        if fs.list:
-            # fs.list can be None when there's nothing to parse
-            for field in fs.list:
-                if field.filename:
-                    obj.add(field.name, field)
-                else:
-                    obj.add(field.name, field.value)
+        # fs.list can be None when there's nothing to parse
+        for field in fs.list or ():
+            if field.filename:
+                obj.add(field.name, field)
+            else:
+                obj.add(field.name, field.value)
         return obj
 
-    from_fieldstorage = classmethod(from_fieldstorage)
-
     def __getitem__(self, key):
         for k, v in reversed(self._items):
             if k == key:
@@ -136,16 +130,12 @@ class MultiDict(DictMixin):
 
     def dict_of_lists(self):
         """
-        Returns a dictionary where each key is associated with a
-        list of values.
+        Returns a dictionary where each key is associated with a list of values.
         """
-        result = {}
-        for key, value in self.iteritems():
-            if key in result:
-                result[key].append(value)
-            else:
-                result[key] = [value]
-        return result
+        r = {}
+        for key, val in self.iteritems():
+            r.setdefault(key, []).append(val)
+        return r
 
     def __delitem__(self, key):
         items = self._items
@@ -195,7 +185,7 @@ class MultiDict(DictMixin):
     def popitem(self):
         return self._items.pop()
 
-    def update(self, other=None, **kwargs):
+    def extend(self, other=None, **kwargs):
         if other is None:
             pass
         elif hasattr(other, 'items'):
@@ -292,10 +282,12 @@ class UnicodeMultiDict(DictMixin):
             # decode FieldStorage's field name and filename
             value = copy.copy(value)
             if self.decode_keys:
-                value.name = value.name.decode(self.encoding, self.errors)
+                if not isinstance(value.name, unicode):
+                    value.name = value.name.decode(self.encoding, self.errors)
             if value.filename:
-                value.filename = value.filename.decode(self.encoding,
-                                                       self.errors)
+                if not isinstance(value.filename, unicode):
+                    value.filename = value.filename.decode(self.encoding,
+                                                           self.errors)
         elif not isinstance(value, unicode):
             try:
                 value = value.decode(self.encoding, self.errors)
@@ -423,11 +415,54 @@ class UnicodeMultiDict(DictMixin):
 
 _dummy = object()
 
+class TrackableMultiDict(MultiDict):
+    tracker = None
+    name = None
+    def __init__(self, *args, **kw):
+        if '__tracker' in kw:
+            self.tracker = kw.pop('__tracker')
+        if '__name' in kw:
+            self.name = kw.pop('__name')
+        MultiDict.__init__(self, *args, **kw)
+    def __setitem__(self, key, value):
+        MultiDict.__setitem__(self, key, value)
+        self.tracker(self, key, value)
+    def add(self, key, value):
+        MultiDict.add(self, key, value)
+        self.tracker(self, key, value)
+    def __delitem__(self, key):
+        MultiDict.__delitem__(self, key)
+        self.tracker(self, key)
+    def clear(self):
+        MultiDict.clear(self)
+        self.tracker(self)
+    def setdefault(self, key, default=None):
+        result = MultiDict.setdefault(self, key, default)
+        self.tracker(self, key, result)
+        return result
+    def pop(self, key, *args):
+        result = MultiDict.pop(self, key, *args)
+        self.tracker(self, key)
+        return result
+    def popitem(self):
+        result = MultiDict.popitem(self)
+        self.tracker(self)
+        return result
+    def update(self, other=None, **kwargs):
+        MultiDict.update(self, other, **kwargs)
+        self.tracker(self)
+    def __repr__(self):
+        items = ', '.join(['(%r, %r)' % v for v in self.iteritems()])
+        return '%s([%s])' % (self.name or self.__class__.__name__, items)
+    def copy(self):
+        # Copies shouldn't be tracked
+        return MultiDict(self)
+
 class NestedMultiDict(MultiDict):
     """
     Wraps several MultiDict objects, treating it as one large MultiDict
     """
-    
+
     def __init__(self, *dicts):
         self.dicts = dicts
 
@@ -508,7 +543,7 @@ class NestedMultiDict(MultiDict):
                 yield key
 
     iterkeys = __iter__
-    
+
 class NoVars(object):
     """
     Represents no variables; used when no variables
@@ -516,7 +551,7 @@ class NoVars(object):
 
     This is read-only
     """
-    
+
     def __init__(self, reason=None):
         self.reason = reason or 'N/A'
 
@@ -575,32 +610,3 @@ class NoVars(object):
     iteritems = iterkeys
     values = keys
     itervalues = iterkeys
-
-__test__ = {
-    'general': """
-    >>> d = MultiDict(a=1, b=2)
-    >>> d['a']
-    1
-    >>> d.getall('c')
-    []
-    >>> d.add('a', 2)
-    >>> d['a']
-    2
-    >>> d.getall('a')
-    [1, 2]
-    >>> d['b'] = 4
-    >>> d.getall('b')
-    [4]
-    >>> d.keys()
-    ['a', 'a', 'b']
-    >>> d.items()
-    [('a', 1), ('a', 2), ('b', 4)]
-    >>> d.mixed()
-    {'a': [1, 2], 'b': 4}
-    >>> MultiDict([('a', 'b')], c=2)
-    MultiDict([('a', 'b'), ('c', 2)])
-    """}
-
-if __name__ == '__main__':
-    import doctest
-    doctest.testmod()
